;; /* Copyright (c) SRI International 2003, 2004. */
;;;;;;;;;;;;;;;;;;;;;;;;;;* -*- Mode: Lisp -*- ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; vim: syntax=lisp
;; decide3.2.lisp --
;; Author          : Ashish Tiwari
;; Created On      : ?? 
;; Last Modified By: Ashish Tiwari
;; Last Modified On: Wed Apr 26, 2006
;; Status          : Unknown, use with caution
;;
;; HISTORY :
;; CSL 2005 paper implementation
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(provide 'gb)		;; This module is named GB: GB based dp for reals
(cl:defpackage "gb"
  (:nicknames "GB")
  (:export "saturate" "sos" "set-debug-level" "set-newU" "set-newV" "get-newU" "get-newV")
  (:use "polynomial-representation-core" "cl" "user" "clos"))

(in-package "gb")	;; The core procedure

(defstruct POL pol wit c)	;; polynomial and witness and side condition c


(defvar *newU* nil)	;; new vars U > 0
(defvar *newV* nil)	;; new vars V >= 0
(defvar *dlevel* nil)		;; Default t. Set and Get using accessors

;; ---------------------------------------------------------------------------
;; Introducing new PARAMETERS to bound the GB computation
;; ---------------------------------------------------------------------------
;; External
(defvar *degreeratio* 2)	;; generated-poly-degree/input-poly-degree
(defvar *lengthratio* 5)	;; generated-poly-length/input-poly-length
;; Internal
(defvar *degreebound* 1)	;; degree of polynomial
(defvar *lengthbound* 1)	;; number of monomials in any polynomial
;; external functions to set parameters
(defun set-degree-ratio (n) (setf *degreeratio* n))
(defun set-length-ratio (n) (setf *lengthratio* n))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; Print function
;; ---------------------------------------------------------------------------
(defun pp (E &key (key #'(lambda(x) x)))
  (loop for i in E collect (format nil "~a,  " (prep:polyrepPrint (funcall key i)))))

(defmacro print-debug (level &rest others)
  `(if (and *dlevel* (> ,level *dlevel*))
       (funcall #'format ,@others)))

(defun set-debug-level (n)
  (setf *dlevel* n))

(defmacro test-and-return (funname R st)
  `(if (eq ,st 'inconsistent) (return-from ,funname (values nil ,R))
       (multiple-value-bind (newR wit) (consistencyCheck ,R)
         (if (eq newR 'inconsistent) (return-from ,funname (values nil wit)))
         newR)))

(defun set-newU (u) (setf *newU* u))
(defun set-newV (u) (setf *newV* u))
(defun get-newU () *newU*)
(defun get-newV () *newV*)
;; ----------------------------------------------------------------------------

;; ----------------------------------------------------------------------------
;; Consistency Checks. Input could be inconsistent (chainPartialAllWith)
;; ----------------------------------------------------------------------------
(defun consistencyCheck (plist)
  ;;(if (eq plist 'inconsistent) (return-from consistencyCheck 'inconsistent))
  (loop for i in plist 
	for j = (formula-status i)
	if (eq j 'satisfiable) collect i
	if (eq j 'inconsistent) return (values j (POL-wit i))))

;; formula-status : check if the atomic formula is not 0 > 0
;; Return value: 'valid, 'satisfiable, 'inconsistent
(defun formula-status (p)
  (let ((poly (POL-pol p)))
  (cond ((null poly) 'valid)
	((prep:polyrepConstant? poly)
	 (cond ((eq (caar poly) 0) 'valid)
	       (t 'inconsistent)))
	((null (sos-test (list p))) 'inconsistent)
	(t 'satisfiable))))
;; ----------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
(defun saturate (E R &optional (S nil) (ERS nil) (Eold nil))
  (cond ((not (or E R S ERS)) (values t Eold))
	((and (or E R S) (null ERS)) (sos E R S Eold))
	((or E R S)
	 (multiple-value-bind (st Enew) (sos E R S Eold)
	   (if st (saturate nil nil nil ERS Enew) (values st Enew))))
	(t
      	 (let ((p (car ERS)) (op (cadr ERS)))
	   (multiple-value-bind (st Enew)
	     (case op (= (sos (list p) nil nil Eold))
		      (> (sos nil (list p) nil Eold))
		      (>= (sos nil nil (list p) Eold))
		      (t (format t "Unidentified operator~A~%" op) (break)))
	     (if st (saturate nil nil nil (cddr ERS) Enew) (values st Enew)))))))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; CSL paper procedure
;; Input: list of pols: E=0, R>0, S>=0 and Context Eold
;; Output: (status, witness or newGB)
;; ---------------------------------------------------------------------------
(defun sos (E R &optional (S nil) (Eold nil))
  (declare (special *newU* *newV*))
  (init-params E R S)
  (print-debug 5 t "Input equations E:~%~a~% " (pp E))
  (print-debug 5 t "Input inequalities R: ~%~a~% " (pp R))
  (print-debug 5 t "Input inequalities S: ~%~a~% " (pp S))
  (let ((origU *newU*) (origV *newV*)
	(origX (prep:get-variables)) (origC (prep:get-parameters))
	(E (mapcar #'make-pole E)) (R (mapcar #'make-polr R))
	(S (mapcar #'make-pols S)))
    (let ((newE (add-extension-vars E R S))
	  (tempU *newU*) (tempV *newV*)
	  (tempX (prep:get-variables)) (tempC (prep:get-parameters)))
      (multiple-value-bind (st EE) (sos-att newE nil nil Eold 0 nil)
	(if st 
	    (progn (setf *newU* tempU) (setf *newV* tempV) 
		   (prep:set-variables tempX) (prep:set-parameters tempC)
		   (print-debug 5 t "SATISFIABLE~%")
		   (values t EE))
	    (progn (setf *newU* origU) (setf *newV* origV) 
		   (prep:set-variables origX) (prep:set-parameters origC)
		   (print-debug 7 t "UNSATISFIABLE WITNESS IS ~%~a~%" (pp (get-witness EE) :key #'POL-pol))
		   (values st EE)))))))

;; E, R, S, Eold: 'POL list; K: int;
(defun sos-att (E R S Eold K Eret)
  (if (> K 13) (return-from sos-att (values t Eret)))
  (multiple-value-bind (st gb1 gb2) 
	(grobner (if (eq K 0) E (add-extension-vars E R S)) Eold)
    (if (null st) (return-from sos-att (values st gb1)))
    (if st (print-debug 5 t "printing new GB:~%~a~%~a~%" 
		(pp gb1 :key #'POL-pol) (pp gb2 :key #'POL-pol)))
    (multiple-value-bind (st ngb) (sos-test gb1)
      (if (null st) (values st ngb)
	  (let ((gb12 (append gb1 gb2)))
	    (sos-test2 gb12 K (if (eq K 0) gb12 Eret)))))))

;; Initialize/Set the parameters, if necessary
(defun maxdegree (E)
  (if (null E) 0 (apply #'max (mapcar #'prep:polyrepTotalDegreePoly E))))
(defun maxlength (E)
  (if (null E) 0 (apply #'max (mapcar #'length E))))
(defun init-params (E R S)
  (declare (special *degreebound* *degreeration* *lengthbound* *lengthratio*))
  (let ((d (max (maxdegree E) (maxdegree R) (maxdegree S)))
	(n (max (maxlength E) (maxlength R) (maxlength S))))
    (setf *degreebound* (max (* *degreeratio* d) *degreebound*))
    (setf *lengthbound* (max (* *lengthratio* n) *lengthbound*))))
(defun bound-check? (p)
  (declare (special *degreebound* *lengthbound*))
  (and (<= (prep:polyrepTotalDegreePoly p) *degreebound*)
       (<= (length p) *lengthbound*)))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; Detection Phase
;; ---------------------------------------------------------------------------
(defun sos-test (gb)
  (let ((p (some #'pn1? gb)))
    (if p (print-debug 6 t "Poly ~a found to be inconsistent." 
		(prep:polyrepPrint (POL-pol p))))
    (if p (values nil (POL-wit p)) 
	(values t gb))))
  
;; return value of 0 should be treated differently.
;; when pn-poly returns 0, then we only know that each monomial of p = 0
;; BUT currently, we will just say that p is inconsistent!!!
(defun pn1? (p)
  (if (eq (pn-poly? (POL-pol p)) 1) p))

(defun pn-poly? (p)
  (cond ((every #'(lambda(x) (> (car x) 0)) p)
         (p-poly?* p))
	((every #'(lambda(x) (< (car x) 0)) p)
         (p-poly?* p))
	(t nil)))

;; ans = 1 if all mu's in p >= 0 and one mu > 0; ans = 0 if p >= 0, ans = nil otherwise
(defun p-poly?* (p &optional (ans 0))
  (if (null p) (return-from p-poly?* ans))
  (let ((sgn (p-mono? (cdar p))))
    (cond ((eq sgn -1) nil)
	  ((eq sgn 1) (p-poly?* (cdr p) 1))
	  (t (p-poly?* (cdr p) ans)))))

;; ans = 1 if mu > 0, ans = 0 if mu >= 0, ans = -1 otherwise
(defun p-mono? (mu &optional (ans 1))
  (cond ((null mu) ans)
        ((var-in-list? (caar mu) *newU*) (p-mono? (cdr mu) ans))
        ((var-in-list? (caar mu) *newV*) (p-mono? (cdr mu) 0))
	((evenp (cdar mu)) (p-mono? (cdr mu) 0))
        (t -1)))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; The second phase: nontrivial detection algorithm.
;; Algo: First we remove all equations whose leading PP is an ORIG variable.
;; ---------------------------------------------------------------------------
(defun sos-test2 (gb K Eret)
  (let ((rgb (loop for i in gb if (not (useless? (POL-pol i))) collect i)))
    (print-debug 5 t "Second phase: testing unsatisfiability of:~%~a~%" (pp rgb :key #'POL-pol))
    (if (<= (length rgb) 1) (values t gb)
        (let* ((rgb (sort rgb #'lesser?))
	       (rgb (if (eq K 0) (homogenize rgb) rgb)))
    	  (multiple-value-bind (st wit) (sos-test2* rgb K Eret)
	  	(if st (values t gb) (values nil wit)))))))

;; rename leading PP of "p" to a new var to make it smaller; and recurse
(defun sos-test2* (gb K Eret)
  (if (or (> K 13) (<= (length gb) 1)) (return-from sos-test2* (values t Eret)))
  (let* ((gb1 (cycle gb K))
	 (p (some #'almost-positive? gb1))
	 (p (if p p (some #'(lambda(x) (almost-positive? x 2)) gb1)))
	 (p (if p p (maxOfList gb #'> :key #'(lambda(x)(length (POL-pol x))) :ret #'POL-pol)))
	 (pol (make-pol (list (cons 1 (cdar p))) -1 nil))) ;; NOTE: Ignoring U,V distinction
    (print-debug 2 t "Main Poly is:~%~a~%" (pp (list p)))
    (sos-att nil (list pol) nil gb (+ K 1) Eret)))	;; NOTE: Ignoring U,V distinction/witness

(defun cycle (l n)
  (let ((N (length l)))
    (loop for i from 0 upto (- N 1) collect (nth (mod (+ n i) N) l))))

(defun homogenize (gb)
  (let ((newu (get-new-var '>)))
    (loop for i in gb collect (homogenizePoly i newu))))

(defun homogenizePoly (p u)
  (let ((N (prep:polyrepTotalDegreePoly (POL-pol p))))
    (make-POL :pol
    (loop for i in (POL-pol p) collect
	  (let ((n (prep:polyrepTotalDegreePoly (list i))))
	    (if (> N n) 
	        (append i (list (cons u (- N n))))
	        i))) :wit (POL-wit p) :c (POL-c p))))

(defun useless? (p)
  (or (is-orig-pp? (cdar p)) (binomial? p)	;; COMPLETE
      (not-positive-pp? (cdar p))))		;; INCOMPLETE

;; check also for order of first two monomials?
(defun binomial? (p)
  (or (null (cddr p)) 	;; binomial
      (and (null (cddar p)) (eq (cdadar p) 1) 
	   (or (and (> (caar p) 0) (every #'(lambda(x) (< (car x) 0)) (cdr p)))
      	       (and (< (caar p) 0) (every #'(lambda(x) (> (car x) 0)) (cdr p)))))))

;; atmost N positive (or negative) monomials in p
;; INPUT: POL, OUTPUT: polyrep poly
(defun almost-positive? (P &optional (N 1) (n 0) (m 0))
  (let ((p (POL-pol P)))
  ;; if nonlinear, return nil
  (if (or (cddar p) (> (cdadar p) 1)) (return-from almost-positive? nil))
  (loop for i in p do (if (> (car i) 0) (setf n (+ n 1))) (if (> n N) (return nil)))
  (if (<= n N) p
      (progn
        (loop for i in p do (if (< (car i) 0) (setf m (+ m 1))) (if (> m N) (return nil)))
	(if (<= m N) p nil)))))

(defun is-orig-pp? (mu)
  (and (null (cdr mu)) (eq (cdar mu) 1) (is-orig-var? (caar mu))))
(defun is-orig-var? (v)
  (not (var-in-list? v (append *newV* *newU*))))

(defun var-in-list? (var vlist)
  (member var vlist :test #'prep:var-equal?))

(defun not-positive-pp? (mu)
  (some #'(lambda(x) (is-orig-var? (car x))) mu))
;; ---------------------------------------------------------------------------
  
;; ---------------------------------------------------------------------------
;; Grobner Basis Computation
;; ---------------------------------------------------------------------------
(defun grobner (E &optional (Eold nil))
  (print-debug 5 t "Order for constructing GB is ~a~%" (prep:get-order))
  (print-debug 5 t "Constructing GB for:~%~a~%~%" (pp E :key #'POL-pol))
  (multiple-value-bind (st Enew Eold) (gb-full E Eold)
    (if st (print-debug 5 t "Eqns sat. GB = ~%~a~%" (pp (append Enew Eold) :key #'POL-pol)))
    (if st 
	(values t (remove-if #'null Enew :key #'POL-pol) Eold)
	(values nil Enew))))

(defun make-pole (p) (make-pol p nil '=))
(defun make-polr (p) (make-pol p nil '> ))
(defun make-pols (p) (make-pol p nil '>=))
(defun make-pol (p wit c)
  (make-POL :pol p :wit wit :c c))		;; Constraint is OP for R,S
;; ----------------------------------------------------------------------------

;; ----------------------------------------------------------------------------
;; gb1-normalize( unproc_list, gb) --> GB(gb+unproc_list)
;; Algo: (i) pick first p from unproc_list; normalize it by gb 
;;      (ii) collapse gb by p: (newgb, oldgb) <-- gb2-collapse( p, gb )
;;     (iii) superpose oldgb with p: unproc_new <-- gb3-superpose( p, gbold)
;;      (iv) gb1-full( unproc_new+newgb, gbold+p )
;; ----------------------------------------------------------------------------
(defun gb-full (unproc_list &optional (gbold nil) (gbnew nil))
  (cond ((null unproc_list) (values t gbnew gbold))
	((and (null gbold) (null gbnew))
	 (gb-full (cdr unproc_list) nil (list (car unproc_list))))
	(t
	 (let ((plist (list (car unproc_list))))
	 (multiple-value-bind (st newE oldE)	;; collapse new eqn
		(dynamicSimplifyRbyErec (append gbold gbnew) plist)
	   (print-debug 2 t "Simplified new equation ~a by GB ~a~%" 
			 (pp plist :key #'POL-pol) (pp (append gbold gbnew) :key #'POL-pol))
	   (let* ((newE  (test-and-return gb-full newE st))
		  (plist (nconc newE oldE)))
	     (if (car plist)
	         (multiple-value-bind (st newGBOld oldGBOld)
				(dynamicSimplifyRbyErec plist gbold)
		   (show-debug "old" plist gbold newGBOld oldGBOld)
		   (let ((newGBOld (test-and-return gb-full newGBOld st)))
	             (multiple-value-bind (st newGBNew oldGBNew)
				(dynamicSimplifyRbyErec plist gbnew)
			(show-debug "new" plist gbnew newGBNew oldGBNew)
		   	(let ((newGBNew (test-and-return gb-full newGBNew st))
			      (newCPs (allPairCPs (car plist) (append oldGBOld oldGBNew))))
			  (print-debug 4 t "No of new CPS = ~a~%" (length newCPs))
			  (print-debug 4 t "~a~%" (pp newCPs :key #'POL-pol))
			  (gb-full (nconc newCPs newGBOld newGBNew (cdr unproc_list)) 
					    oldGBOld (nconc plist oldGBNew))))))
	       	 (gb-full (cdr unproc_list) gbold gbnew))))))))

(defun allPairCPs (p qlist)
  (delete nil (mapcar #'(lambda(x) (superposition p x)) qlist)))

(defun show-debug (which plist a b c)
	   	       (print-debug 2 t "-------~%")
	   	       (print-debug 2 t "Simplified ~a GB by new equation~a~%" which
			 (pp plist :key #'POL-pol))
	   	       (print-debug 2 t "Old GB:~a~%" (pp a :key #'POL-pol)) 
	   	       (print-debug 2 t "Changed:~a~%" (pp b :key #'POL-pol))
	   	       (print-debug 2 t "UnChanged:~a~%" (pp c :key #'POL-pol))
	   	       (print-debug 2 t "-------~%"))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; Introducing extension/slack variables.
;; E,R,S are sorted s.t. x > 0 occurs first. This way we don't incur additional
;; overhead when we introduce new var for x > 0 cases (no reordering reqd too)
;; return-value, E, R, S: 'POL list
;; ---------------------------------------------------------------------------
(defun add-extension-vars (E R S)
  (let* ((newR (sort R #'lesser?))
	 (newS (sort S #'lesser?))
	 (newE (sort E #'lesser?))
	 (E2E (mapcar #'(lambda(x) (make-pol (POL-pol x) (list x) nil)) newE))
	 (R2E (mapcar #'(lambda(x) (addU x '>)) newR))
	 (S2E (mapcar #'(lambda(x) (addU x '>=)) newS)))
    (merge 'list (merge 'list E2E R2E #'lesser?) S2E #'lesser?)))

(defun lesser? (x y)
  (case (prep:polyrepComparePoly (POL-pol x) (POL-pol y))
	(prep:less t) (t nil)))

(defun get-new-var (op)
  (let* ((newu (intern (gensym))))
    (if (eq op '>) 
	(setf *newU* (cons newu *newU*))
	(setf *newV* (cons newu *newV*)))
    (prep:set-parameters (append (prep:get-parameters) (list newu)))
    (print-debug 4 t "new order is ~a~%" (prep:get-order))
    newu))
  
;; P, ret: POL; copy witness from P to ret.
(defun addU (P op)
  (let ((newu (get-new-var op))
	(pol (POL-pol P)) 
	(wit (if (eq (POL-wit P) -1) nil (list P))))
    (make-pol (prep:polyrepAddPoly pol (list (list -1 (cons newu 1)))) wit nil)))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; Stuff from decide2.0.lisp; that is required here.
;; Normalize R using the equations E as rewrite rules
;; Return a (status, newR, oldR). 
;; ----------------------------------------------------------------------------
(defun dynamicSimplifyRbyErec (E R)
  (dynamicSimplifyRbyErec* E R nil nil))

(defun dynamicSimplifyRbyErec* (E R new old)
  (if (null R) (values 'satisfiable new old)
      (multiple-value-bind (changed? newp) (fully-reduce E (car R))
	(if changed? 
	    (dynamicSimplifyRbyErec* E (cdr R) (cons newp new) old)
	    (dynamicSimplifyRbyErec* E (cdr R) new (cons (car R) old))))))

;; normalize p by E
(defun fully-reduce (E p)
  (fully-reduce* E p nil 0 nil (length E)))

(defun fully-reduce* (E p Eold n changed? N)
  (cond ((eq n N) (values changed? p))
	((null E) (fully-reduce* (nreverse Eold) p nil n changed? N))
	(t
	 (multiple-value-bind (newp oldp) (dynamicSimplify (car E) (list p))
	   (declare (ignore oldp))
	   (if newp (fully-reduce* E (car newp) Eold 0 t N)
	       (fully-reduce* (cdr E) p (cons (car E) Eold) (+ n 1) changed? N))))))

;; Dynamic variants: Simplify each poly in plist by p = 0
;; One step parallel rewrite on each poly in plist by p = 0
(defun dynamicSimplify (p plist)
  (cond ((null plist)
	 (values nil nil))
	((null p)
	 (values nil plist))
	(t (dynamicSimplify* p plist))))

;; Simplify each poly in plist by p = 0
(defun dynamicSimplify* (p plist)
  (let* ((c0s0 (car (POL-pol p)))
	 (c0 (* (car c0s0) -1))		;; -1 for moving cisi to RHS!!
	 (cisi (prep:polyrepDividePolyCst (cdr (POL-pol p)) c0))
	 (s0 (cdr c0s0)))
    (dynamicSimplifyList* s0 cisi plist (POL-wit p))))	;; pass the WITNESS

;; Replace s0 by cisi in each poly in plist: 
;; Return only (changed old) pair
(defun dynamicSimplifyList* (s0 cisi plist wit &optional (new nil) (old nil))
  (if (null plist) (values new old)
      (let* ((p (car plist))
	     (newp (simplify-poly* s0 cisi (POL-pol p))))
        (if (not (eq newp 0)) (print-debug 1 t "Simplified~A to~% ~A~%" p newp))
	(if (eq newp 0)
	    (dynamicSimplifyList* s0 cisi (cdr plist) wit new (cons p old))
	    (let* ((newp1 (normalizePoly newp))
		   (newp2 (combine-wit wit p))
		   (newp3 (make-POL :pol newp1 :wit newp2)))
	      (dynamicSimplifyList* s0 cisi (cdr plist) wit (cons newp3 new) old))))))

;; Replace s0 by cisi in each poly in plist
(defun simplify-poly* (s0 cisi poly)
  (if (null poly) 0
      (let ((poly1 (simplify-mono* s0 cisi (car poly)))
	    (poly2 (simplify-poly* s0 cisi (cdr poly))))
        (cond ((and (eq poly1 0) (eq poly2 0))
	       0)
	      ((eq poly1 0)
	       (cons (car poly) poly2))	;; ordering guarantees this
	      ((eq poly2 0)
	       (prep:polyrepAddPoly poly1 (cdr poly)))
	      (t 
	       (prep:polyrepAddPoly poly1 poly2))))))

(defun simplify-mono* (s0 cisi mono)
  (let* ((d0 (car mono))
	 (t0 (cdr mono))
	 (newt0 (simplify-pp* s0 cisi t0)))
    (if (eq newt0 0)
	0
	(prep:polyrepMultiplyCstPoly d0 newt0))))

(defun simplify-pp* (s0 cisi t0)
  (let ((t0bys0 (prep:polyrepDividePP t0 s0)))
    (if (eq t0bys0 0)
	0
	(prep:polyrepMultiplyMonoPoly (cons 1 t0bys0) cisi))))
(defun normalizePoly (p)
  (if (null p) p
      (let ((c0 (abs (caar p))))
	(if (eq c0 1) p (prep:polyrepDividePolyCst p c0)))))
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; Witness functions
;; ---------------------------------------------------------------------------
(defun combine-wit (p q)
  (union (if (POL-p p) (POL-wit p) p) (POL-wit q)))

(defun get-witness (E)
  (print-debug 2 t "Get-witness function called with WITNESS~%~A~%" E)
  E)
;; ---------------------------------------------------------------------------

;; ---------------------------------------------------------------------------
;; Superposition: Critical pair computation
;; ---------------------------------------------------------------------------
(defun superposition (p q)
  (let* ((cp1 (superposition* p q))
	 (cp (if (bound-check? cp1) cp1 nil)))
    (if (null cp) nil 
        (make-POL :pol cp :wit (combine-wit p q) :c nil))))

;; removing null checks on p,q
(defun superposition* (p1 q1)
  (let* ((p (POL-pol p1))
	 (q (POL-pol q1))
	 (c0s0 (car p))
	 (d0t0 (car q))
	 (c0 (car c0s0))
	 (d0 (car d0t0))
	 (d0p c0)
	 (c0p (- 0 d0)))
    (multiple-value-bind (overlap? s0p t0p)
	(overlap-pp-pp (cdr c0s0) (cdr d0t0) (POL-c p1) (POL-c q1))
	(if overlap?
	    (let ((ans (prep:polyrepAddPoly 
			(prep:polyrepMultiplyMonoPoly (cons c0p s0p) (cdr p)) 
			(prep:polyrepMultiplyMonoPoly (cons d0p t0p) (cdr q)))))
	      (setf (POL-c q1) (cons t0p (POL-c q1)))
	      (normalizePoly ans))
	    nil))))

(defun overlap-pp-pp (mu nu nolistmu nolistnu)
  (let ((order (prep:get-order)))
    (multiple-value-bind (st mu0 nu0) (overlap-pp-pp* mu nu order nil nil nil)
      (if (or (member mu0 nolistmu :test #'multiple?)
	      (member nu0 nolistnu :test #'multiple?))
	  (progn (print-debug 5 t "OPTIMIZATION-1 works~%") nil)
	  (values st mu0 nu0)))))

(defun overlap-pp-pp* (mu nu order yes mu1 nu1)
  (when (and (null yes) (member (car order) (append *newU* *newV*) :test #'prep:var-equal?))
      (print-debug 5 t "OPTIMIZATION-2 works~%")
      (return-from overlap-pp-pp* nil))
  (if (null order) 
      (values yes (nreverse mu1) (nreverse nu1))
      (let* ((x1 (car order))
	     (x1c1 (assoc x1 mu :test #'prep:var-equal?))
	     (x1c2 (assoc x1 nu :test #'prep:var-equal?))
	     (c1 (cdr x1c1))
	     (c2 (cdr x1c2)))
	(cond ((and (null c1) (null c2))
	       (overlap-pp-pp* mu nu (cdr order) yes mu1 nu1))
	      ((null c1)
	       (overlap-pp-pp* mu nu (cdr order) yes (cons x1c2 mu1) nu1))
	      ((null c2)
	       (overlap-pp-pp* mu nu (cdr order) yes mu1 (cons x1c1 nu1)))
	      ((= c1 c2)
	       (overlap-pp-pp* mu nu (cdr order) t mu1 nu1))
	      ((> c1 c2)
	       (overlap-pp-pp* mu nu (cdr order) t mu1
			(cons (cons (car x1c2) (- c1 c2)) nu1)))
	      ((< c1 c2)
	       (overlap-pp-pp* mu nu (cdr order) t 
			(cons (cons (car x1c2) (- c2 c1)) mu1) nu1))))))

(defun multiple? (mu nu)
  (cond ((null nu) t)
	((null mu) nil)
	((prep:var-equal? (caar nu) (caar mu))
	 (if (<= (cdar nu) (cdar mu))
	     (multiple? (cdr mu) (cdr nu)) 
	     nil))
	(t (multiple? (cdr mu) nu))))
;; ---------------------------------------------------------------------------

;; ----------------------------------------------------------------------------
;; maxOfList: return maximum element of list, use fn to compare.
;; ----------------------------------------------------------------------------
(defun maxOfList (plist fn &key (key #'(lambda(x) x)) (ret #'(lambda(x) x)))
  (assert (car plist))
  (maxOfList* (cdr plist) fn (car plist) key ret))

(defun maxOfList* (plist fn ans key ret)
  (if (null plist) (funcall ret ans)
      (if (funcall fn (funcall key (car plist)) (funcall key ans))
	  (maxOfList* (cdr plist) fn (car plist) key ret)
	  (maxOfList* (cdr plist) fn ans key ret))))
;; ----------------------------------------------------------------------------

